package uk.ac.bath.oy213.ip.server;

import uk.ac.bath.oy213.ip.server.commands.net.*;
import uk.ac.bath.oy213.ip.server.commands.user.*;
import uk.ac.bath.oy213.ip.server.location.LocationTable;
import uk.ac.bath.oy213.ip.server.net.ServerListenerThread;
import uk.ac.bath.oy213.ip.server.security.TokenRegistry;
import uk.ac.bath.oy213.ip.server.util.Colour;
import uk.ac.bath.oy213.ip.server.util.DatabaseConnection;
import uk.ac.bath.oy213.ip.server.util.FileHandler;
import uk.ac.bath.oy213.ip.server.util.LoggingLevel;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * Created by Ollie on 28/03/15.
 */
public class Server {

    private static Server instance;

    private DatabaseConnection databaseConnection;
    private LocationTable locationTable;

    private Scanner sc;
    private ArrayList<NetCommand> netCommands;
    private ArrayList<UserCommand> userCommands;
    private ServerListenerThread listenerThread;
    private TokenRegistry tokenRegistry;
    private boolean running;

    private final int PORT = 25565;

    public static void main(String[] args) {

        Colour.assignColours();

        new Server();

    }

    public static Server instance() {

        return instance;

    }

    public Server() {

        instance = this;

        netCommands = new ArrayList<>();
        userCommands = new ArrayList<>();

        loadCommands();

        long startTime = System.currentTimeMillis();

        System.out.println("###################" + Colour.ANSI_BLUE + " E-biz Server " + Colour.ANSI_RESET + "###################\n");
        System.out.println("                    " + Colour.ANSI_GREEN + "Version 1.01" + Colour.ANSI_RESET);
        System.out.println("                      " + Colour.ANSI_GREEN + "Group: 3" + Colour.ANSI_RESET + "\n");
        System.out.println("################ " + Colour.ANSI_BLUE + "Integrated Project" + Colour.ANSI_RESET + " ################\n");

        locationTable = new LocationTable();

        log(LoggingLevel.SERVER, "Establishing connection to database...");

        databaseConnection = new DatabaseConnection("localhost");

        log(LoggingLevel.SERVER, "Enabling listener threads...");

        tokenRegistry = new TokenRegistry();

        listenerThread = new ServerListenerThread(PORT);
        listenerThread.start();

        log(LoggingLevel.SERVER, "Server listening on address: " + Colour.ANSI_GREEN + getPublicIP() + Colour.ANSI_RESET + ".");
        log(LoggingLevel.SERVER, "Server listening on port: "+Colour.ANSI_GREEN+PORT+Colour.ANSI_RESET+".");

        long finishTime = System.currentTimeMillis();
        long startupTime = finishTime - startTime;

        log(LoggingLevel.SERVER, "Ready! Completed in " + TimeUnit.MILLISECONDS.toSeconds(startupTime) + "s (" + startupTime + "ms).\n");

        log(LoggingLevel.INFO, "Type '" + Colour.ANSI_GREEN + "stop" + LoggingLevel.INFO.getReset() + "' to exit.");
        log(LoggingLevel.INFO, "Type '" + Colour.ANSI_GREEN + "help" + LoggingLevel.INFO.getReset() + "' for a list of further commands.");

        prompt();

        sc = new Scanner(System.in);

        startListening();

    }

    public void startListening() {

        String input;
        String[] inputArray;
        running = true;

        try {

            //loop until exit is executed
            while (running) {

                input = sc.nextLine();

                inputArray = input.split(" ");

                //if the command is not found, notify the user
                if (!parseCommand(inputArray[0], Arrays.copyOfRange(inputArray, 1, inputArray.length))) {

                    log(LoggingLevel.WARNING, "Command '" + Colour.ANSI_GREEN + inputArray[0] + LoggingLevel.WARNING.getReset() + "' not found, type '" + Colour.ANSI_GREEN + "help" + LoggingLevel.WARNING.getReset() + "' for more information.");

                }

                //user input prompt
                if (running) prompt();

            }

            exit();

        } catch (NoSuchElementException e) { //EOF - quit

            System.out.println("\n");
            exit();

        }

    }

    public boolean parseCommand(String cmd, String[] args) {

        System.out.print("\n");

        for (UserCommand c: userCommands) {

            if (c.hasAlias(cmd)) {

                c.dispatch(args);
                return true;

            }

        }

        return false;

    }

    public void stopListening() {

        this.running = false;

    }

    private String getPublicIP() {

        try {

            BufferedReader in = new BufferedReader(new InputStreamReader(new URL("http://checkip.amazonaws.com").openStream()));
            String ip = in.readLine();

            in.close();

            return ip;

        } catch (Exception e) {

            return "N/A";

        }

    }

    public void loadCommands() {

        userCommands.add(new StopCommand());
        userCommands.add(new HelpCommand());
        userCommands.add(new ListAuthTableCommand());
        userCommands.add(new ListConnectionsCommand());
        userCommands.add(new SendPacketCommand());
        userCommands.add(new TerminateConnectionCommand());

        netCommands.add(new AuthenticateUserCommand());
        netCommands.add(new CreateUserCommand());
        netCommands.add(new GetLocalCardsCommand());
        netCommands.add(new LogoutCommand());
        netCommands.add(new RemoveCardCommand());
        netCommands.add(new RequestCardCommand());
        netCommands.add(new RequestOwnCards());
        netCommands.add(new RequestSavedCardsCommand());
        netCommands.add(new RequestUsersMasterCardCommand());
        netCommands.add(new SaveCardCommand());
        netCommands.add(new SetMasterCardCommand());
        netCommands.add(new ShareCardCommand());
        netCommands.add(new StopSharingCardCommand());
        netCommands.add(new UpdateAccountDetailsCommand());
        netCommands.add(new UpdateCardCommand());
        netCommands.add(new UpdateLocationCommand());
        netCommands.add(new UploadCardCommand());

    }

    public void prompt() {

        System.out.print("\n" + Colour.ANSI_RESET);

    }

    public void log(LoggingLevel level, String message) {

        level.getStream().println(level.getValue() + message + Colour.ANSI_RESET);

    }

    public ArrayList<UserCommand> getUserCommands() {

        return userCommands;

    }

    public ArrayList<NetCommand> getNetCommands() {

        return netCommands;

    }

    public ServerListenerThread getListenerThread() {

        return listenerThread;

    }

    public DatabaseConnection getDatabaseConnection() {

        return databaseConnection;

    }

    public TokenRegistry getTokenRegistry() {

        return tokenRegistry;

    }

    public LocationTable getLocationTable() {

        return locationTable;

    }

    public void exit() {

        log(LoggingLevel.SERVER, "Stopping server...");

        log(LoggingLevel.SERVER, "Shutting down listener threads...");
        listenerThread.shutDown();
        sc.close();

        log(LoggingLevel.SERVER, "Server stopped.\n");

        System.exit(0);

    }

}
